part of '../../ngx_core.dart';

/// The injection token providing the session backend.
const OpaqueToken<Map<String, String>> sessionBackend =
    OpaqueToken<Map<String, String>>('session.backend');

/// The interface that should be implemented by services managing user identification.
abstract class IdentityService {
  /// The number of seconds in which the user will be logged out automatically if he remains inactive.
  int get authTimeout;

  /// The URL of the login page.
  String get loginUrl;

  /// Authenticates a user by the given [username] and [password].
  /// Returns the authenticated user identity, or 'null' if such an identity cannot be found.
  Future<Identity> authenticate(String username, String password);

  /// Creates a new identity from the specified [map] in JSON format.
  Identity create(Map<String, dynamic> map);

  /// Invalidates the authentication data associated to the specified [identity].
  Future<void> invalidate(Identity identity);
}

/// Manages the user sessions.
class Session implements OnDestroy {
  /// Creates a new session service.
  Session(@Optional() @sessionBackend Map<String, String> backend, this._app,
      this._identityService, this._router)
      : _backend = backend ?? dom.window.localStorage {
    try {
      _loginFromBackend();
    } on SessionException {
      switchIdentity(null);
    }
    _subscription = _router.onNavigationStart.listen((_) => _renewAuthStatus());
  }

  /// The application instance.
  final Application _app;

  /// The backend used to persist session data.
  final Map<String, String> _backend;

  /// The service managing user identification.
  final IdentityService _identityService;

  /// The handler of "identity" events.
  final BehaviorSubject<Identity> _onIdentity =
      BehaviorSubject<Identity>.seeded(null);

  /// The routing service.
  final Router _router;

  /// The subscription to the router events.
  StreamSubscription<String> _subscription;

  /// The identity object associated with the currently logged-in user.
  Identity get identity => _onIdentity.value;

  set identity(Identity value) {
    if (value != _onIdentity.value) _onIdentity.add(value);
  }

  /// Value indicating whether the current user is a guest (not authenticated).
  bool get isGuest => identity == null;

  /// The URL of the login page.
  String get loginUrl => _identityService.loginUrl;

  /// The stream of "identity" events.
  Stream<Identity> get onIdentity => _onIdentity.stream;

  /// The URL that the user should be redirected to after login.
  String get returnUrl =>
      dom.window.sessionStorage['${_app.id}:session.returnUrl'] ?? '/';
  set returnUrl(String value) =>
      dom.window.sessionStorage['${_app.id}:session.returnUrl'] = value;

  /// The access token associated to this session.
  SessionToken token = SessionToken();

  /// Logs in a user by the specified [username] and [password].
  /// Returns the new identity after the user is logged in.
  ///
  /// Throws a [SessionException] if the user could not be authenticated.
  Future<Identity> login(String username, String password) async {
    final identity = await _identityService.authenticate(username, password);
    if (identity == null)
      throw const SessionException(
          'Could not find an identity corresponding to the specified username and password.');
    switchIdentity(identity);
    return identity;
  }

  /// Redirects the user browser to the login page.
  void loginRequired() {
    switchIdentity(null);
    returnUrl = _router.current.toUrl();
    _router.navigateByUrl(loginUrl);
  }

  /// Logs out the current user and removes authentication-related cookie data.
  Future<void> logout() async {
    await _identityService.invalidate(identity);
    switchIdentity(null);
  }

  /// Method invoked before the service is destroyed.
  @override
  void ngOnDestroy() => _subscription.cancel();

  /// Switches to a new [identity] for the current user.
  ///
  /// If the [identity] is `null`, it means switching the current user to be a guest.
  /// The [token] is ignored when [identity] is `null`.
  ///
  /// Throws a [SessionException] if the specified identity or session token is invalid.
  void switchIdentity(Identity identity, [SessionToken token]) {
    if (identity != null) {
      if (identity.id is int && identity.id <= 0)
        throw const SessionException('The user identity is invalid.');
      else if (identity.id is String && identity.id.isEmpty)
        throw const SessionException('The user identity is invalid.');

      token ??= SessionToken(identity.authKey,
          DateTime.now().add(Duration(seconds: _identityService.authTimeout)));
      if (!token.isValid)
        throw const SessionException('The session token is invalid.');

      _backend['${_app.id}:session.token'] = jsonEncode(token);
      _backend['${_app.id}:session.identity'] = jsonEncode(identity);

      this.token = token;
      this.identity = identity;
    } else {
      _backend
        ..remove('${_app.id}:session.token')
        ..remove('${_app.id}:session.identity');
      this.token = SessionToken();
      this.identity = null;
    }
  }

  /// Retrieves a value from the session backend.
  /// Returns the fetched value, or `null` if the session backend does not contain the specified [key].
  dynamic _getFromBackend(String key) {
    if (!_backend.containsKey(key)) return null;
    try {
      return jsonDecode(_backend[key]);
    } on FormatException {
      return null;
    }
  }

  /// Logs in a user from the session backend.
  /// Returns the new identity after the user is logged in.
  Identity _loginFromBackend() {
    var data = _getFromBackend('${_app.id}:session.identity');
    final identity =
        _identityService.create(data is Map ? data : const <String, dynamic>{});

    data = _getFromBackend('${_app.id}:session.token');
    final token =
        SessionToken.fromJson(data is Map ? data : const <String, dynamic>{});

    switchIdentity(identity, token);
    return identity;
  }

  /// Updates the authentication status according to the session timer.
  void _renewAuthStatus() {
    if (!token.isValid)
      switchIdentity(null);
    else {
      token.expires =
          DateTime.now().add(Duration(seconds: _identityService.authTimeout));
      _backend['${_app.id}:session.token'] = jsonEncode(token);
    }
  }
}
